/*++

Copyright (c) 2002-2003 Alexandr A. Telyatnikov (aka Alter)

Module Name:

    id_probe.cpp

Abstract:

    This is the miniport driver for ATAPI IDE controllers
    With Busmaster DMA support

Author:

    Alex A. Telyatnikov (Alter)

Environment:

    kernel mode only

Notes:

Revision History:

--*/

extern "C" {

#include <ntddk.h>
//#include "miniport.h"

};
#include "stddef.h"

#include "atapi.h"               // includes scsi.h
#include "ntdddisk.h"
#include "ntddscsi.h"
#include "bsmaster.h"

UCHAR         pciBuffer[256];
PBUSMASTER_CONTROLLER_INFORMATION BMList = NULL;
ULONG         BMListLen = 0;
ULONG         maxPciBus = 16;

// debug

ULONG
AtapiEnumBusMasterController__(
    IN PVOID HwDeviceExtension,
    IN PVOID Context,
    IN PVOID BusInformation,
    IN PCHAR ArgumentString,
    IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    OUT PBOOLEAN Again
    );

VOID
AtapiEnumBusMasterController(
    IN PVOID DriverObject,
    PVOID Argument2
    )
{
    HW_INITIALIZATION_DATA hwInitializationData;

    // Zero out structure.
    AtapiZeroMemory(((PCHAR)&hwInitializationData), sizeof(HW_INITIALIZATION_DATA));

    // Set size of hwInitializationData.
    hwInitializationData.HwInitializationDataSize =
      sizeof(HW_INITIALIZATION_DATA);

    // Set entry points.
    hwInitializationData.HwInitialize = AtapiHwInitialize;
    hwInitializationData.HwResetBus = AtapiResetController;
    hwInitializationData.HwStartIo = AtapiStartIo;
    hwInitializationData.HwInterrupt = AtapiInterrupt;

    // Specify size of extensions.
    hwInitializationData.DeviceExtensionSize = sizeof(HW_DEVICE_EXTENSION);
    hwInitializationData.SpecificLuExtensionSize = sizeof(HW_LU_EXTENSION);

    // Indicate PIO device.
    hwInitializationData.MapBuffers = TRUE;

    hwInitializationData.HwFindAdapter = AtapiEnumBusMasterController__;
    hwInitializationData.NumberOfAccessRanges = 4;
    hwInitializationData.AdapterInterfaceType = Isa;

    KdPrint2((PRINT_PREFIX "\n\nATAPI IDE Look for PCI Controllers\n"));
    ScsiPortInitialize(DriverObject,
                       Argument2,
                       &hwInitializationData,
                       NULL);
}


ULONG
AtapiEnumBusMasterController__(
    IN PVOID HwDeviceExtension,
    IN PVOID Context,
    IN PVOID BusInformation,
    IN PCHAR ArgumentString,
    IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    OUT PBOOLEAN Again
    )
{
    PCI_SLOT_NUMBER       slotData;
    PCI_COMMON_CONFIG     pciData;
    ULONG                 busNumber;
    ULONG                 slotNumber;
    ULONG                 funcNumber;
    BOOLEAN               no_buses = FALSE;
    ULONG                 busDataRead;

    UCHAR                 vendorString[5];
    UCHAR                 deviceString[5];
    PUCHAR                vendorStrPtr;
    PUCHAR                deviceStrPtr;

    UCHAR   BaseClass;                  // (ro)
    UCHAR   SubClass;                   // (ro)
    ULONG   VendorID;
    ULONG   DeviceID;
    ULONG   dev_id;

    ULONG   i;
    ULONG   pass=0;

    BOOLEAN found;
    BOOLEAN known;

    vendorStrPtr = vendorString;
    deviceStrPtr = deviceString;
    slotData.u.AsULONG = 0;

    for(pass=0; pass<2; pass++) {
        for(busNumber=0 ;busNumber<maxPciBus && !no_buses; busNumber++) {
            for(slotNumber=0; slotNumber<PCI_MAX_DEVICES  && !no_buses; slotNumber++) {
            for(funcNumber=0; funcNumber<PCI_MAX_FUNCTION && !no_buses; funcNumber++) {
                
//                KdPrint2((PRINT_PREFIX "-- BusID: %x:%x:%x\n",busNumber,slotNumber,funcNumber));
                slotData.u.bits.DeviceNumber   = slotNumber;
                slotData.u.bits.FunctionNumber = funcNumber;

                busDataRead = ScsiPortGetBusData(HwDeviceExtension,
                                            PCIConfiguration, busNumber, slotData.u.AsULONG, 
                                            &pciData, PCI_COMMON_HDR_LENGTH);
                // no more buses
                if(!busDataRead) {
                    no_buses = TRUE;
                    maxPciBus = busNumber;
                    break;
                }
                // no device in this slot
                if(busDataRead == 2)
                    continue;

                if(busDataRead < PCI_COMMON_HDR_LENGTH)
                    continue;

                VendorID  = pciData.VendorID;
                DeviceID  = pciData.DeviceID;
                BaseClass = pciData.BaseClass;
                SubClass  = pciData.SubClass;
                dev_id = VendorID | (DeviceID << 16);
    //            KdPrint2((PRINT_PREFIX "DevId = %8.8X Class = %4.4X/%4.4X\n", dev_id, BaseClass, SubClass ));

                if(BaseClass != PCI_DEV_CLASS_STORAGE)
                    continue;

                KdPrint2((PRINT_PREFIX "Storage Class\n"));
                // look for known chipsets
                found = FALSE;
                known = FALSE;
                for(i=0; i<NUM_BUSMASTER_ADAPTERS; i++) {
                    if(VendorID == BusMasterAdapters[i].nVendorId &&
                       DeviceID == BusMasterAdapters[i].nDeviceId) {
                        KdPrint2((PRINT_PREFIX "Known device\n"));
                        found = TRUE;
                        known = TRUE;
                        break;
                    }
                }

                switch(dev_id) {
                /* additional checks for some supported chipsets */
                case 0xc6931080:
                    if (SubClass != PCI_DEV_SUBCLASS_IDE)
                        found = FALSE;
                    break;

                /* unknown chipsets, try generic DMA if it seems possible */
                default:
                    KdPrint2((PRINT_PREFIX "Default device\n"));
                    if (BaseClass == PCI_DEV_CLASS_STORAGE &&
                        SubClass == PCI_DEV_SUBCLASS_IDE) {
                        found = TRUE;
                    }
                    break;
                }

                if(found) {

                    if(pass) {
                        // fill list of detected devices
                        // it'll be used for further init
                        KdPrint2((PRINT_PREFIX "found suitable device\n"));
                        PBUSMASTER_CONTROLLER_INFORMATION newBMListPtr = BMList+BMListLen;

                        if(known) {
                            RtlCopyMemory(newBMListPtr, (PVOID)&(BusMasterAdapters[i]), sizeof(BUSMASTER_CONTROLLER_INFORMATION));
                        } else {
                            AtapiHexToString(VendorID, (PCHAR*)(&vendorStrPtr));
                            AtapiHexToString(DeviceID, (PCHAR*)(&deviceStrPtr));

                            RtlCopyMemory(&(newBMListPtr->VendorIdStr), (PCHAR)vendorStrPtr, 4);
                            RtlCopyMemory(&(newBMListPtr->DeviceIdStr), (PCHAR)deviceStrPtr, 4);
                            
                            newBMListPtr->nVendorId = VendorID;
                            newBMListPtr->VendorId = (PCHAR)&(newBMListPtr->VendorIdStr);
                            newBMListPtr->VendorIdLength = 4;
                            newBMListPtr->nDeviceId = DeviceID;
                            newBMListPtr->DeviceId = (PCHAR)&(newBMListPtr->DeviceIdStr);
                            newBMListPtr->DeviceIdLength = 4;
                        }
                        newBMListPtr->slotNumber = slotData.u.AsULONG;
                        newBMListPtr->MasterDev = IsMasterDev(&pciData) ? 1 : 0;
                    }

                    BMListLen++;
                }
            }
            }
        }
        if(!pass) {
             if(!BMListLen)
                 break;
             BMList = (PBUSMASTER_CONTROLLER_INFORMATION)ExAllocatePool(NonPagedPool, 
                                         (BMListLen+1)*sizeof(BUSMASTER_CONTROLLER_INFORMATION));
             no_buses = FALSE;
             BMListLen=0;
        }
    }
    return(SP_RETURN_NOT_FOUND);
} // end AtapiEnumBusMasterController()

/*ULONG
AtapiGetDeviceBase(
    ) {
}*/


ULONG 
ScsiPortGetBusDataByOffset(
    IN PVOID  HwDeviceExtension,
    IN BUS_DATA_TYPE  BusDataType,
    IN ULONG  BusNumber,
    IN ULONG  SlotNumber,
    IN PVOID  Buffer,
    IN ULONG  Offset,
    IN ULONG  Length
    )
{
    UCHAR tmp[256];
    ULONG busDataRead;

    if(Offset+Length > 256)
        return 0;

    busDataRead = ScsiPortGetBusData(HwDeviceExtension,
                                     BusDataType,
                                     BusNumber,
                                     SlotNumber,
                                     &tmp,
                                     Offset+Length);
    if(busDataRead < Offset+Length) {
        if(busDataRead < Offset)
            return 0;
        return (Offset+Length-busDataRead);
    }
    RtlCopyMemory(Buffer, tmp+Offset, Length);
    return Length;
}

ULONG
AtapiFindDev(
    IN PVOID  HwDeviceExtension,
    IN BUS_DATA_TYPE  BusDataType,
    IN ULONG  BusNumber,
    IN ULONG  SlotNumber,
    IN ULONG  dev_id,
    IN ULONG  RevID
    )
{
    PCI_COMMON_CONFIG     pciData;
    ULONG                 funcNumber;
    ULONG                 busDataRead;

    ULONG   VendorID;
    ULONG   DeviceID;
    PCI_SLOT_NUMBER       slotData;

    slotData.u.AsULONG = SlotNumber;
    for(funcNumber = 0; funcNumber < PCI_MAX_FUNCTION; funcNumber++) {
        
        slotData.u.bits.FunctionNumber = funcNumber;
        if(slotData.u.AsULONG == SlotNumber)
            continue;

        busDataRead = ScsiPortGetBusData(HwDeviceExtension,
                                         PCIConfiguration,
                                         BusNumber,
                                         slotData.u.AsULONG,
                                         &pciData,
                                         PCI_COMMON_HDR_LENGTH);

        if (busDataRead < PCI_COMMON_HDR_LENGTH) {
            continue;
        }

        VendorID  = pciData.VendorID;
        DeviceID  = pciData.DeviceID;

        if(dev_id != (VendorID | (DeviceID << 16)) )
            continue;
        if(RevID == pciData.RevisionID)
            return 1;
    }
    return 0;
}

/*++

Routine Description:

    This function is called by the OS-specific port driver after
    the necessary storage has been allocated, to gather information
    about the adapter's configuration.

Arguments:

    HwDeviceExtension - HBA miniport driver's adapter data storage
    Context - Address of adapter count
    BusInformation -
    ArgumentString - Used to determine whether driver is client of ntldr or crash dump utility.
    ConfigInfo - Configuration information structure describing HBA
    Again - Indicates search for adapters to continue

Return Value:

    ULONG

--*/
ULONG
AtapiFindBusMasterController(
    IN PVOID HwDeviceExtension,
    IN PVOID Context,
    IN PVOID BusInformation,
    IN PCHAR ArgumentString,
    IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    OUT PBOOLEAN Again
    )
{
    PHW_DEVICE_EXTENSION  deviceExtension = (PHW_DEVICE_EXTENSION)HwDeviceExtension;
    PCI_COMMON_CONFIG     pciData;
    ULONG                 slotNumber;
    ULONG                 busDataRead;

    UCHAR                 vendorString[5];
    UCHAR                 deviceString[5];
    PUCHAR                vendorStrPtr;
    PUCHAR                deviceStrPtr;

    UCHAR   BaseClass;
    UCHAR   SubClass;
    ULONG   VendorID;
    ULONG   DeviceID;
    ULONG   RevID;
    ULONG   dev_id;
    PCI_SLOT_NUMBER       slotData;

    ULONG   i = (ULONG)Context;
    ULONG   channel = BMList[i].channel;
    ULONG   c = 0;
    PUCHAR  ioSpace;
    UCHAR   statusByte;
//    ULONG   irq;

    BOOLEAN found;
    BOOLEAN MasterDev;

    SCSI_PHYSICAL_ADDRESS IoBasePort1;
    SCSI_PHYSICAL_ADDRESS IoBasePort2;


    *Again = FALSE;

    if (!deviceExtension) {
        return SP_RETURN_ERROR;
    }

    vendorStrPtr = vendorString;
    deviceStrPtr = deviceString;

    slotNumber = BMList[i].slotNumber;

    KdPrint2((PRINT_PREFIX "IoBusNumber=%x\n",ConfigInfo->SystemIoBusNumber));
    KdPrint2((PRINT_PREFIX "slotNumber=%x\n",slotNumber));

    busDataRead = ScsiPortGetBusData(HwDeviceExtension,
                                     PCIConfiguration,
                                     ConfigInfo->SystemIoBusNumber,
                                     slotNumber,
                                     &pciData,
                                     PCI_COMMON_HDR_LENGTH);

    if (busDataRead < PCI_COMMON_HDR_LENGTH) {
        return SP_RETURN_ERROR;
    }

    KdPrint2((PRINT_PREFIX "busDataRead\n"));
    if (pciData.VendorID == PCI_INVALID_VENDORID) {
        KdPrint2((PRINT_PREFIX "PCI_INVALID_VENDORID\n"));
        return SP_RETURN_ERROR;
    }

    VendorID  = pciData.VendorID;
    DeviceID  = pciData.DeviceID;
    BaseClass = pciData.BaseClass;
    SubClass  = pciData.SubClass;
    RevID     = pciData.RevisionID;
    dev_id = VendorID | (DeviceID << 16);
    slotData.u.AsULONG = slotNumber;
    KdPrint2((PRINT_PREFIX "DevId = %8.8X Class = %4.4X/%4.4X\n", dev_id, BaseClass, SubClass ));

    if(BaseClass != PCI_DEV_CLASS_STORAGE)
        return(SP_RETURN_NOT_FOUND);

    KdPrint2((PRINT_PREFIX "Storage Class\n"));

    // look for known chipsets
    if(VendorID != BMList[i].nVendorId ||
       DeviceID != BMList[i].nDeviceId) {
        KdPrint2((PRINT_PREFIX "device not suitable\n"));
        return(SP_RETURN_NOT_FOUND);
    }

    found = TRUE;
    switch(dev_id) {
    /* additional checks for some supported chipsets */
    case 0xc6931080:
        if (SubClass != PCI_DEV_SUBCLASS_IDE)
            found = FALSE;
        break;

    /* unknown chipsets, try generic DMA if it seems possible */
    default:
        KdPrint2((PRINT_PREFIX "Default device\n"));
        if (found)
            break;
        if (BaseClass != PCI_DEV_CLASS_STORAGE ||
            SubClass != PCI_DEV_SUBCLASS_IDE) {
            found = FALSE;
        }
        break;
    }

    if(!found)
        return(SP_RETURN_NOT_FOUND);

    KdPrint2((PRINT_PREFIX "found suitable device\n"));

    /***********************************************************/
    /***********************************************************/
    /***********************************************************/

    deviceExtension->NumberChannels = 1;
    MasterDev = IsMasterDev(&pciData);

    if(MasterDev) {
        if((channel==0) && ConfigInfo->AtdiskPrimaryClaimed)
            return(SP_RETURN_NOT_FOUND);
        if((channel==1) && ConfigInfo->AtdiskSecondaryClaimed)
            return(SP_RETURN_NOT_FOUND);
    }

/*    if(!MasterDev) {
        IoBasePort1 = (*ConfigInfo->AccessRanges)[2].RangeStart;

        if ( !MasterDev &&
             ScsiPortConvertPhysicalAddressToUlong(IoBasePort1) )
            deviceExtension->NumberChannels = 2;
    }*/

    deviceExtension->BusMaster = FALSE;
    if(IsBusMaster(&pciData)) {
        if(ConfigInfo->NumberOfAccessRanges >= 5)
            deviceExtension->BaseIoAddressBM_0 =
            // Get the system physical address for this IO range.
                (PIDE_BUSMASTER_REGISTERS)ScsiPortGetDeviceBase(HwDeviceExtension,
                                            ConfigInfo->AdapterInterfaceType,
                                            ConfigInfo->SystemIoBusNumber,
                                            (*ConfigInfo->AccessRanges)[4].RangeStart,
                                            ATA_BMIOSIZE,
                                            TRUE);
        if(deviceExtension->BaseIoAddressBM_0) {
            deviceExtension->BusMaster = TRUE;
        } else
        if(deviceExtension->BaseIoAddressBM_0 = (PIDE_BUSMASTER_REGISTERS)(pciData.u.type0.BaseAddresses[4])) {
            deviceExtension->BusMaster = TRUE;
        }
    }

    if(deviceExtension->BusMaster && !MasterDev) {
        statusByte = ScsiPortReadPortUchar((PUCHAR)&(deviceExtension->BaseIoAddressBM_0->Status));
        if(statusByte & BM_STATUS_SIMPLEX_ONLY) {
            deviceExtension->NumberChannels = 2;
        }
    }
    if((channel > 0) &&
       (deviceExtension->NumberChannels > 1))
        return(SP_RETURN_NOT_FOUND);
    
    /* do extra chipset specific setups */
    switch (dev_id) {
    case 0x522910b9: /* Aladdin need to activate the ATAPI FIFO */
        ChangePciConfig1(0x53, ((a & ~0x01) | 0x02));
	break;

    case 0x4d38105a: /* Promise 66 & 100 need their clock changed */
    case 0x4d30105a:
    case 0x0d30105a:
        if(deviceExtension->BaseIoAddressBM_0) {
            statusByte = ScsiPortReadPortUchar((PUCHAR)(deviceExtension->BaseIoAddressBM_0) + 0x11);
            ScsiPortWritePortUchar((PUCHAR)(deviceExtension->BaseIoAddressBM_0) + 0x11, statusByte | 0x0a);
        }
	/* FALLTHROUGH */

    case 0x4d33105a: /* Promise (all) need burst mode to be turned on */
        if(deviceExtension->BaseIoAddressBM_0) {
            statusByte = ScsiPortReadPortUchar((PUCHAR)(deviceExtension->BaseIoAddressBM_0) + 0x1f);
            ScsiPortWritePortUchar((PUCHAR)(deviceExtension->BaseIoAddressBM_0) + 0x1f, statusByte | 0x01);
	}
	break;

    case 0x00041103: /* HighPoint */
	switch (RevID) {
	case 0x00:
	case 0x01:
	    /* turn off interrupt prediction */
            ChangePciConfig1(0x51, (a & ~0x80));
	    break;

	case 0x02:
	case 0x03:
	case 0x04:
	    /* turn off interrupt prediction */
            ChangePciConfig1(0x51, (a & ~0x02));
            ChangePciConfig1(0x55, (a & ~0x02));
	    /* turn on interrupts */
            ChangePciConfig1(0x5a, (a & ~0x10));

	}
	break;

    case 0x05711106:
    case 0x74091022: /* VIA 82C586, 82C596, 82C686 & AMD 756 default setup */

	/* set prefetch, postwrite */
        ChangePciConfig1(0x41, (a | 0xf0));

	/* set fifo configuration half'n'half */
        ChangePciConfig1(0x43, ((a & 0x90) | 0x2a));

	/* set status register read retry */
        ChangePciConfig1(0x44, (a | 0x08));

	/* set DMA read & end-of-sector fifo flush */
        ChangePciConfig1(0x46, ((a & 0x0c) | 0xf0));

	/* set sector size */
	SetPciConfig2(0x60, DEV_BSIZE);
	SetPciConfig2(0x68, DEV_BSIZE);
	
	/* prepare for ATA-66 on the 82C686 and rev 0x12 and newer 82C596's */
	if (AtapiFindDev(HwDeviceExtension,
                         PCIConfiguration,
                         ConfigInfo->SystemIoBusNumber, 
                         slotNumber, 0x06861106, 0) || 
	    AtapiFindDev(HwDeviceExtension,
                         PCIConfiguration,
                         ConfigInfo->SystemIoBusNumber, 
                         slotNumber, 0x05961106, 0x12)) {
            ChangePciConfig4(0x50, (a | 0x070f070f));
	}
	break;

    case 0x10001042:	/* RZ 100? known bad, no DMA */
    case 0x10011042:
    case 0x06401095:	/* CMD 640 known bad, no DMA */
    case 0x06011039:
	deviceExtension->BaseIoAddressBM_0 = 0;
        deviceExtension->BusMaster = FALSE;
    }

    /*
     * the Cypress chip is a mess, it contains two ATA functions, but 
     * both channels are visible on the first one.
     * simply ignore the second function for now, as the right
     * solution (ignoring the second channel on the first function)
     * doesn't work with the crappy ATA interrupt setup on the alpha.
     */
    if (dev_id == 0xc6931080 && slotData.u.bits.FunctionNumber > 1) {
        goto exit_findbm;
    }

    // Indicate number of buses.
    ConfigInfo->NumberOfBuses = 1;
    if(!ConfigInfo->InitiatorBusId[0]) {
        ConfigInfo->InitiatorBusId[0] = (CHAR)(IoGetConfigurationInformation()->ScsiPortCount);
/*        if(ConfigInfo->NumberOfBuses > 1)
            ConfigInfo->InitiatorBusId[1] = (CHAR)(IoGetConfigurationInformation()->ScsiPortCount)+1;*/
    }
    // Indicate four devices can be attached to the adapter
    ConfigInfo->MaximumNumberOfTargets = (UCHAR)(deviceExtension->NumberChannels * 2);
    ConfigInfo->MaximumTransferLength = 0x10000;
    // Save the Interrupe Mode for later use
    deviceExtension->DevID = dev_id;
    if (MasterDev) {
        if (ConfigInfo->AdapterInterfaceType == MicroChannel) {
            ConfigInfo->InterruptMode = LevelSensitive;
        } else {
            ConfigInfo->InterruptMode = Latched;
        }
        ConfigInfo->BusInterruptLevel = (channel == 0 ? 14 : 15);
    } else {
        ConfigInfo->SlotNumber = BMList[i].slotNumber;
        /* primary and secondary channels share the same interrupt */
        if(!ConfigInfo->BusInterruptVector || 
            (ConfigInfo->BusInterruptVector != pciData.u.type0.InterruptLine)) {
            ConfigInfo->BusInterruptVector = pciData.u.type0.InterruptLine;  // set default value
            if(!ConfigInfo->BusInterruptVector)
                ConfigInfo->BusInterruptVector = 10;
        }
    }
    deviceExtension->InterruptMode = ConfigInfo->InterruptMode;

    found = FALSE;

    for (; channel < (BMList[i].channel + deviceExtension->NumberChannels); channel++, c++) {

        // Check if the range specified if not used by another driver
        if(MasterDev) {
            (*ConfigInfo->AccessRanges)[channel * 2 + 0].RangeStart =
                ScsiPortConvertUlongToPhysicalAddress(channel ? IO_WD2 : IO_WD1);
            (*ConfigInfo->AccessRanges)[channel * 2 + 0].RangeLength = ATA_IOSIZE;

            (*ConfigInfo->AccessRanges)[channel * 2 + 1].RangeStart =
                ScsiPortConvertUlongToPhysicalAddress((channel ? IO_WD2 : IO_WD1) + ATA_ALTOFFSET);
            (*ConfigInfo->AccessRanges)[channel * 2 + 1].RangeLength = ATA_ALTIOSIZE;
        }

        if(!ScsiPortValidateRange(HwDeviceExtension,
                                  ConfigInfo->AdapterInterfaceType,
                                  ConfigInfo->SystemIoBusNumber,
                                  (*ConfigInfo->AccessRanges)[channel * 2 + 0].RangeStart,
                                  ATA_IOSIZE,
                                  TRUE) )
            continue;
        
        if(!ScsiPortValidateRange(HwDeviceExtension,
                                  ConfigInfo->AdapterInterfaceType,
                                  ConfigInfo->SystemIoBusNumber,
                                  (*ConfigInfo->AccessRanges)[channel * 2 + 1].RangeStart,
                                  ATA_ALTIOSIZE,
                                  TRUE) )
            continue;

        // Ok, translate adresses to io-space
        IoBasePort1 = (*ConfigInfo->AccessRanges)[channel * 2 + 0].RangeStart;
        IoBasePort2 = (*ConfigInfo->AccessRanges)[channel * 2 + 1].RangeStart;
        
        if(ScsiPortConvertPhysicalAddressToUlong(IoBasePort2)) {
            if(!(MasterDev /* || USE_16_BIT */)) {
                IoBasePort2 = ScsiPortConvertUlongToPhysicalAddress(
                                  ScsiPortConvertPhysicalAddressToUlong(IoBasePort2) + 2);
            }
        } else {
            IoBasePort2 = ScsiPortConvertUlongToPhysicalAddress(
                              ScsiPortConvertPhysicalAddressToUlong(IoBasePort1) + ATA_PCCARD_ALTOFFSET);
        }

        // Get the system physical address for this IO range.
        ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                        ConfigInfo->AdapterInterfaceType,
                                        ConfigInfo->SystemIoBusNumber,
                                        IoBasePort1,
                                        ATA_IOSIZE,
                                        TRUE);

        // Check if ioSpace accessible.
        if (!ioSpace) {
            continue;
        }

        if(deviceExtension->BusMaster) {
            deviceExtension->BaseIoAddressBM[c] =
                deviceExtension->BaseIoAddressBM_0 + (channel ? ATA_BM_OFFSET1 : 0);
        }

        // Check master.
        ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->DriveSelect, 
                               IDE_DRIVE_SELECT | IDE_DRIVE_1);
        ScsiPortStallExecution(1);
        GetBaseStatus((PIDE_REGISTERS_1)ioSpace, statusByte);
        if ((statusByte & 0xf8) == 0xf8 || 
            (statusByte == 0xa5)) {
            // Check slave.
            ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->DriveSelect, 
                                   IDE_DRIVE_SELECT | IDE_DRIVE_2);
            ScsiPortStallExecution(1);
            GetBaseStatus((PIDE_REGISTERS_1)ioSpace, statusByte);
            if ((statusByte & 0xf8) == 0xf8 || 
                (statusByte == 0xa5)) {
                // No controller at this base address.
free_iospace_1:
                // Release all allocated resources
                ScsiPortFreeDeviceBase(HwDeviceExtension,
                                       ioSpace);
                continue;
            }
        }
        deviceExtension->BaseIoAddress1[c] = (PIDE_REGISTERS_1)(ioSpace);

        // Get the system physical address for the second IO range.
        ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                        ConfigInfo->AdapterInterfaceType,
                                        ConfigInfo->SystemIoBusNumber,
                                        IoBasePort2,
                                        ATA_ALTIOSIZE,
                                        TRUE);

        deviceExtension->BaseIoAddress2[c] = (PIDE_REGISTERS_2)(ioSpace);
        if(!deviceExtension->BaseIoAddress2[c]) {
            // Release all allocated resources
            ioSpace = (PUCHAR)(deviceExtension->BaseIoAddress1[c]);
            goto free_iospace_1;
        }

        // Search for devices on this controller.
        if (FindDevices(HwDeviceExtension,
                    FALSE,
                    c)) {
            found = TRUE;
        }

        deviceExtension->PrimaryAddress = FALSE;
        // Claim primary or secondary ATA IO range.
        if (ScsiPortConvertPhysicalAddressToUlong(IoBasePort1) == IO_WD1) {
            ConfigInfo->AtdiskPrimaryClaimed = TRUE;
            deviceExtension->PrimaryAddress = TRUE;

        } else
        if (ScsiPortConvertPhysicalAddressToUlong(IoBasePort2) == IO_WD2) {
            ConfigInfo->AtdiskSecondaryClaimed = TRUE;
        }
    }

    if(deviceExtension->BusMaster) {
        ConfigInfo->NeedPhysicalAddresses = TRUE;
        if(!MasterDev) {
            ConfigInfo->Dma32BitAddresses = TRUE;
        }
//        ConfigInfo->Master = TRUE;
    }

exit_findbm:

    if(!found) {
        if(deviceExtension->BaseIoAddress1[0])
            ScsiPortFreeDeviceBase(HwDeviceExtension,
                                   deviceExtension->BaseIoAddress1[0]);
        if(deviceExtension->BaseIoAddress2[0])
            ScsiPortFreeDeviceBase(HwDeviceExtension,
                                   deviceExtension->BaseIoAddress2[0]);

        if(deviceExtension->BaseIoAddress1[1])
            ScsiPortFreeDeviceBase(HwDeviceExtension,
                                   deviceExtension->BaseIoAddress1[1]);
        if(deviceExtension->BaseIoAddress2[1])
            ScsiPortFreeDeviceBase(HwDeviceExtension,
                                   deviceExtension->BaseIoAddress2[1]);

        if(deviceExtension->BaseIoAddressBM_0)
            ScsiPortFreeDeviceBase(HwDeviceExtension,
                                   deviceExtension->BaseIoAddressBM_0);
        return SP_RETURN_NOT_FOUND;
    }

    return SP_RETURN_FOUND;

} // end AtapiFindBusMasterController()



/*++

Routine Description:

    This function is called by the OS-specific port driver after
    the necessary storage has been allocated, to gather information
    about the adapter's configuration.

Arguments:

    HwDeviceExtension - HBA miniport driver's adapter data storage
    Context - Address of adapter count
    ArgumentString - Used to determine whether driver is client of ntldr or crash dump utility.
    ConfigInfo - Configuration information structure describing HBA
    Again - Indicates search for adapters to continue

Return Value:

    ULONG

--*/
ULONG
AtapiFindController(
    IN PVOID HwDeviceExtension,
    IN PVOID Context,
    IN PVOID BusInformation,
    IN PCHAR ArgumentString,
    IN OUT PPORT_CONFIGURATION_INFORMATION ConfigInfo,
    OUT PBOOLEAN Again
    )
{
    PHW_DEVICE_EXTENSION deviceExtension = (PHW_DEVICE_EXTENSION)HwDeviceExtension;
    PULONG               adapterCount    = (PULONG)Context;
    PUCHAR               ioSpace;
    ULONG                i;
    ULONG                irq;
    ULONG                portBase;
    ULONG                retryCount;
/*    PCI_SLOT_NUMBER      slotData;
    PPCI_COMMON_CONFIG   pciData;*/
//    ULONG                pciBuffer;
    BOOLEAN              atapiOnly;
    UCHAR                statusByte;
    BOOLEAN              preConfig = FALSE;
    //
    // The following table specifies the ports to be checked when searching for
    // an IDE controller.  A zero entry terminates the search.
    CONST ULONG AdapterAddresses[5] = {0x1F0, 0x170, 0x1e8, 0x168, 0};

    // The following table specifies interrupt levels corresponding to the
    // port addresses in the previous table.
    CONST ULONG InterruptLevels[5] = {14, 15, 11, 10, 0};

    if (!deviceExtension) {
        return SP_RETURN_ERROR;
    }
    // Check to see if this is a special configuration environment.
    portBase = irq = 0;
    if (ArgumentString) {

        irq = AtapiParseArgumentString(ArgumentString, "Interrupt");
        if (irq ) {

            // Both parameters must be present to proceed
            portBase = AtapiParseArgumentString(ArgumentString, "BaseAddress");
            if (!portBase) {

                // Try a default search for the part.
                irq = 0;
            }
        }
    }

    // Scan though the adapter address looking for adapters.
    if (ScsiPortConvertPhysicalAddressToUlong((*ConfigInfo->AccessRanges)[0].RangeStart) != 0) {
        ioSpace =  (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                         ConfigInfo->AdapterInterfaceType,
                                         ConfigInfo->SystemIoBusNumber,
                                         (*ConfigInfo->AccessRanges)[0].RangeStart,
                                         (*ConfigInfo->AccessRanges)[0].RangeLength,
                                         (BOOLEAN) !((*ConfigInfo->AccessRanges)[0].RangeInMemory));
        *Again = FALSE;
        // Since we have pre-configured information we only need to go through this loop once
        preConfig = TRUE;
        portBase = ScsiPortConvertPhysicalAddressToUlong((*ConfigInfo->AccessRanges)[0].RangeStart);
    }

    while (AdapterAddresses[*adapterCount] != 0) {

        retryCount = 4;

        for (i = 0; i < 4; i++) {
            // Zero device fields to ensure that if earlier devices were found,
            // but not claimed, the fields are cleared.
            deviceExtension->DeviceFlags[i] &= ~(DFLAGS_ATAPI_DEVICE | DFLAGS_DEVICE_PRESENT | DFLAGS_TAPE_DEVICE);
        }
        // Get the system physical address for this IO range.

        // Check if configInfo has the default information
        // if not, we go and find ourselves
        if (preConfig == FALSE) {

            if (portBase) {
                ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                                ConfigInfo->AdapterInterfaceType,
                                                ConfigInfo->SystemIoBusNumber,
                                                ScsiPortConvertUlongToPhysicalAddress(portBase),
                                                8,
                                                TRUE);
            } else {
                ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                                ConfigInfo->AdapterInterfaceType,
                                                ConfigInfo->SystemIoBusNumber,
                                                ScsiPortConvertUlongToPhysicalAddress(AdapterAddresses[*adapterCount]),
                                                8,
                                                TRUE);
            }

        }// ConfigInfo check

        // Update the adapter count.
        (*adapterCount)++;

        // Check if ioSpace accessible.
        if (!ioSpace) {
            continue;
        }
        // check if Primary/Secondary Master IDE claimed
        if((ioSpace == (PUCHAR)IO_WD1) &&
           (ConfigInfo->AtdiskPrimaryClaimed)) {
            continue;
        } else
        if((ioSpace == (PUCHAR)IO_WD2) &&
           (ConfigInfo->AtdiskSecondaryClaimed)) {
            continue;
        }

retryIdentifier:

        // Select master.
        ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->DriveSelect, IDE_DRIVE_SELECT_1);

        // Check if card at this address.
        ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->CylinderLow, 0xAA);

        // Check if indentifier can be read back.
        if ((statusByte = ScsiPortReadPortUchar(&((PIDE_REGISTERS_1)ioSpace)->CylinderLow)) != 0xAA) {

            KdPrint2((PRINT_PREFIX "AtapiFindController: Identifier read back from Master (%x)\n",
                        statusByte));

            statusByte = ScsiPortReadPortUchar(&((PATAPI_REGISTERS_2)ioSpace)->AlternateStatus);

            if (statusByte & IDE_STATUS_BUSY) {

                i = 0;

                // Could be the TEAC in a thinkpad. Their dos driver puts it in a sleep-mode that
                // warm boots don't clear.
                do {
                    ScsiPortStallExecution(1000);
                    statusByte = ScsiPortReadPortUchar(&((PATAPI_REGISTERS_1)ioSpace)->Command);
                    KdPrint2((PRINT_PREFIX 
                                "AtapiFindController: First access to status %x\n",
                                statusByte));
                } while ((statusByte & IDE_STATUS_BUSY) && ++i < 10);

                if (retryCount-- && (!(statusByte & IDE_STATUS_BUSY))) {
                    goto retryIdentifier;
                }
            }

            // Select slave.
            ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->DriveSelect, 0xB0);

            // See if slave is present.
            ScsiPortWritePortUchar(&((PIDE_REGISTERS_1)ioSpace)->CylinderLow, 0xAA);

            if ((statusByte = ScsiPortReadPortUchar(&((PIDE_REGISTERS_1)ioSpace)->CylinderLow)) != 0xAA) {

                KdPrint2((PRINT_PREFIX 
                            "AtapiFindController: Identifier read back from Slave (%x)\n",
                            statusByte));

                // No controller at this base address.
                ScsiPortFreeDeviceBase(HwDeviceExtension,
                                       ioSpace);
                continue;
            }
        }

        // Record base IO address.
        deviceExtension->BaseIoAddress1[0] = (PIDE_REGISTERS_1)(ioSpace);

        // Fill in the access array information only if default params are not in there.
        if (preConfig == FALSE) {

            // An adapter has been found request another call, only if we didn't get preconfigured info.
            *Again = TRUE;

            if (portBase) {
                (*ConfigInfo->AccessRanges)[0].RangeStart = ScsiPortConvertUlongToPhysicalAddress(portBase);
            } else {
                (*ConfigInfo->AccessRanges)[0].RangeStart =
                    ScsiPortConvertUlongToPhysicalAddress(AdapterAddresses[*adapterCount - 1]);
            }

            (*ConfigInfo->AccessRanges)[0].RangeLength = 8;
            (*ConfigInfo->AccessRanges)[0].RangeInMemory = FALSE;

            // Indicate the interrupt level corresponding to this IO range.
            if (irq) {
                ConfigInfo->BusInterruptLevel = irq;
            } else {
                ConfigInfo->BusInterruptLevel = InterruptLevels[*adapterCount - 1];
            }

            if (ConfigInfo->AdapterInterfaceType == MicroChannel) {
                ConfigInfo->InterruptMode = LevelSensitive;
            } else {
                ConfigInfo->InterruptMode = Latched;
            }
        }

        // Get the system physical address for the second IO range.
        if (portBase) {
            ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                            ConfigInfo->AdapterInterfaceType,
                                            ConfigInfo->SystemIoBusNumber,
                                            ScsiPortConvertUlongToPhysicalAddress(portBase + ATA_ALTOFFSET),
                                            ATA_ALTIOSIZE,
                                            TRUE);
        } else {
            ioSpace = (PUCHAR)ScsiPortGetDeviceBase(HwDeviceExtension,
                                            ConfigInfo->AdapterInterfaceType,
                                            ConfigInfo->SystemIoBusNumber,
                                            ScsiPortConvertUlongToPhysicalAddress(AdapterAddresses[*adapterCount - 1] + ATA_ALTOFFSET),
                                            ATA_ALTIOSIZE,
                                            TRUE);
        }

        deviceExtension->BaseIoAddress2[0] = (PIDE_REGISTERS_2)(ioSpace);

        deviceExtension->NumberChannels = 1;

        ConfigInfo->NumberOfBuses = 1;
        ConfigInfo->MaximumNumberOfTargets = 2;

        // Indicate maximum transfer length is 64k.
        ConfigInfo->MaximumTransferLength = 0x10000;

        KdPrint2((PRINT_PREFIX 
                   "AtapiFindController: Found IDE at %x\n",
                   deviceExtension->BaseIoAddress1[0]));

        // For Daytona, the atdisk driver gets the first shot at the
        // primary and secondary controllers.
        if (preConfig == FALSE) {

            if (*adapterCount - 1 < 2) {

                // Determine whether this driver is being initialized by the
                // system or as a crash dump driver.
                if (ArgumentString) {

                    if (AtapiParseArgumentString(ArgumentString, "dump") == 1) {
                        KdPrint2((PRINT_PREFIX 
                                   "AtapiFindController: Crash dump\n"));
                        atapiOnly = FALSE;
                        deviceExtension->DriverMustPoll = TRUE;
                    } else {
                        KdPrint2((PRINT_PREFIX 
                                   "AtapiFindController: Atapi Only\n"));
                        atapiOnly = TRUE;
                        deviceExtension->DriverMustPoll = FALSE;
                    }
                } else {

                    KdPrint2((PRINT_PREFIX 
                               "AtapiFindController: Atapi Only\n"));
                    atapiOnly = TRUE;
                    deviceExtension->DriverMustPoll = FALSE;
                }

            } else {
                atapiOnly = FALSE;
            }

            // If this is a PCI machine, pick up all devices.
            deviceExtension->DWordIO = FALSE;
/*            pciData = (PPCI_COMMON_CONFIG)&pciBuffer;

            slotData.u.bits.DeviceNumber = 0;
            slotData.u.bits.FunctionNumber = 0;

            if (ScsiPortGetBusData(deviceExtension,
                                   PCIConfiguration,
                                   0,                  // BusNumber
                                   slotData.u.AsULONG,
                                   pciData,
                                   sizeof(pciBuffer))) {

                atapiOnly = FALSE;

                // Wait on doing this, until a reliable method
                // of determining support is found.
    #if 0
                deviceExtension->DWordIO = TRUE;
    #endif

            } else {
                deviceExtension->DWordIO = FALSE;
            }*/

        } else {

            atapiOnly = FALSE;
            deviceExtension->DriverMustPoll = FALSE;

        }// preConfig check

        // Save the Interrupe Mode for later use
        deviceExtension->InterruptMode = ConfigInfo->InterruptMode;

        // Search for devices on this controller.
        if (FindDevices(HwDeviceExtension,
                        atapiOnly,
                        0)) {

            // Claim primary or secondary ATA IO range.
            if (portBase) {
                switch (portBase) {
                case IO_WD2:
                    ConfigInfo->AtdiskSecondaryClaimed = TRUE;
                    deviceExtension->PrimaryAddress = FALSE;
                    break;
                case IO_WD1:
                    ConfigInfo->AtdiskPrimaryClaimed = TRUE;
                    deviceExtension->PrimaryAddress = TRUE;
                    break;
                default:
                    break;
                }
            } else {
                if (*adapterCount == 1) {
                    ConfigInfo->AtdiskPrimaryClaimed = TRUE;
                    deviceExtension->PrimaryAddress = TRUE;
                } else if (*adapterCount == 2) {
                    ConfigInfo->AtdiskSecondaryClaimed = TRUE;
                    deviceExtension->PrimaryAddress = FALSE;
                }
            }

            return(SP_RETURN_FOUND);
        }
    }

    // The entire table has been searched and no adapters have been found.
    // There is no need to call again and the device base can now be freed.
    // Clear the adapter count for the next bus.
    *Again = FALSE;
    *(adapterCount) = 0;

    return(SP_RETURN_NOT_FOUND);

} // end AtapiFindController()
